#!/usr/bin/env python3

###
# Copyright (c) 2002-2005, Jeremiah Fincher
# Copyright (c) 2009, James McCoy
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

from __future__ import print_function

import supybot

import os
import sys
import time
import os.path
import optparse

def error(s):
    sys.stderr.write(textwrap.fill(s))
    sys.stderr.write(os.linesep)
    sys.exit(-1)

import supybot.conf as conf
from supybot.questions import *

copyright = '''
###
# Copyright (c) %s, %%s
# All rights reserved.
#
%%s
###
''' % time.strftime('%Y')
# Here we use strip() instead of lstrip() on purpose.
copyright = copyright.strip()

license = '''
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
'''
license = license.lstrip()

pluginTemplate = '''
%s

from supybot import utils, plugins, ircutils, callbacks
from supybot.commands import *
from supybot.i18n import PluginInternationalization


_ = PluginInternationalization('%s')


class %s(callbacks.Plugin):
    """%s"""
    %s


Class = %s


# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:
'''.lstrip() # This removes the newlines that precede and follow the text.

configTemplate = '''
%s

from supybot import conf, registry
try:
    from supybot.i18n import PluginInternationalization
    _ = PluginInternationalization('%s')
except:
    # Placeholder that allows to run the plugin on a bot
    # without the i18n module
    _ = lambda x: x


def configure(advanced):
    # This will be called by supybot to configure this module.  advanced is
    # a bool that specifies whether the user identified themself as an advanced
    # user or not.  You should effect your configuration by manipulating the
    # registry as appropriate.
    from supybot.questions import expect, anything, something, yn
    conf.registerPlugin(%r, True)


%s = conf.registerPlugin(%r)
# This is where your configuration variables (if any) should go.  For example:
# conf.registerGlobalValue(%s, 'someConfigVariableName',
#     registry.Boolean(False, _("""Help for someConfigVariableName.""")))


# vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:
'''.lstrip()


__init__Template = '''
%s

"""
%s: %s
"""

import sys
import supybot
from supybot import world

# Use this for the version of this plugin.
__version__ = ""

# XXX Replace this with an appropriate author or supybot.Author instance.
__author__ = supybot.authors.unknown

# This is a dictionary mapping supybot.Author instances to lists of
# contributions.
__contributors__ = {}

# This is a url where the most recent plugin package can be downloaded.
__url__ = ''

from . import config
from . import plugin
from importlib import reload
# In case we're being reloaded.
reload(config)
reload(plugin)
# Add more reloads here if you add third-party modules and want them to be
# reloaded when this plugin is reloaded.  Don't forget to import them as well!

if world.testing:
    from . import test

Class = plugin.Class
configure = config.configure


# vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:
'''.lstrip()

pyprojectTemplate = '''
[build-system]
requires = ["setuptools"]

[project]
name = "limnoria-%(name_lowercase)s"
version = "0.0.0"
authors = [
    { name = "%(author_name)s" }
]
readme = "README.md"
dependencies = [
    "limnoria",
]
classifiers = [
    'Environment :: Plugins',
    'Programming Language :: Python :: 3',
    'Topic :: Communications :: Chat',
]

[project.entry-points.'limnoria.plugins']
%(name)s = "limnoria_%(name_lowercase)s"

[tool.setuptools.package-dir]
"limnoria_%(name_lowercase)s" = "."
"limnoria_%(name_lowercase)s.local" = "local/"

[tool.setuptools.package-data]
"limnoria_%(name_lowercase)s" = ["locales/*.po"]
'''.lstrip()

testTemplate = '''
%s

from supybot.test import *


class %sTestCase(PluginTestCase):
    plugins = (%r,)


# vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:
'''.lstrip()

readmeTemplate = '''
%s
'''.lstrip()

def _main():
    global copyright
    global license
    parser = optparse.OptionParser(usage='Usage: %prog [options]',
                                   version='Supybot %s' % conf.version)
    parser.add_option('-n', '--name', action='store', dest='name',
                      help='sets the name for the plugin.')
    parser.add_option('-t', '--thread', action='store_true', dest='threaded',
                      help='makes the plugin threaded.')
    parser.add_option('-a', '--author', '--real-name', action='store',
                      dest='realName', help='Determines who the copyright is '
                      'assigned to.')
    parser.add_option('-d', '--desc', action='store', dest='desc',
                      help='Short description of plugin.')
    (options, args) = parser.parse_args()
    if options.name:
        name = options.name
        threaded = options.threaded
    else:
        name = something('What should the name of the plugin be?')
        if name.endswith('.py'):
            name = name[:-3]
        while name[0].islower():
            print('Plugin names must begin with a capital letter.')
            name = something('What should the name of the plugin be?')
            if name.endswith('.py'):
                name = name[:-3]

        if os.path.exists(name):
            error('A file or directory named %s already exists; remove or '
                  'rename it and run this program again.' % name)

        print(textwrap.fill(textwrap.dedent("""
        Sometimes you'll want a callback to be threaded.  If its methods
        (command or regexp-based, either one) will take a significant amount
        of time to run, you'll want to thread them so they don't block the
        entire bot.""").strip()))
        print()
        threaded = yn('Does your plugin need to be threaded?')

    if options.realName:
        realName = options.realName
    else:
        realName = something(textwrap.dedent("""
        What is your name, so I can fill in the copyright and license
        appropriately?
        """).strip())

        if not yn('Do you wish to use Supybot\'s license for your plugin?'):
            license = '#'

    if not options.desc:
        options.desc = something(textwrap.dedent("""
            Please provide a short description of the plugin:
            """).strip())

    if threaded:
        threaded = 'threaded = True'
    else:
        threaded = 'pass'
    if name.endswith('.py'):
        name = name[:-3]
    while name[0].islower():
        print('Plugin names must begin with a capital letter.')
        name = something('What should the name of the plugin be?')
        if name.endswith('.py'):
            name = name[:-3]
    copyright %= (realName, license)
    pathname = name

    # Make the directory.
    os.mkdir(pathname)

    def writeFile(filename, s):
        fd = open(os.path.join(pathname, filename), 'w')
        try:
            fd.write(s)
        finally:
            fd.close()

    writeFile('plugin.py', pluginTemplate % (copyright, name, name,
                                             options.desc, threaded, name))
    writeFile('config.py', configTemplate % (copyright, name, name, name, name,
                                             name))
    writeFile('__init__.py', __init__Template % (copyright, name, options.desc))
    writeFile('pyproject.toml', pyprojectTemplate % {
        'copyright': copyright,
        'name': name,
        'name_lowercase': name.lower(),
        'author_name': realName.replace(r'"', r'\"'),  # good enough escape
    })
    writeFile('test.py', testTemplate % (copyright, name, name))
    writeFile('README.md', readmeTemplate % (options.desc,))

    pathname = os.path.join(pathname, 'local')
    os.mkdir(pathname)
    writeFile('__init__.py',
              '# Stub so local is a module, used for third-party modules\n')

    print('Your new plugin template is in the %s directory.' % name)

def main():
    try:
        _main()
    except KeyboardInterrupt:
        print()
        output("""It looks like you cancelled out of this script before it was
        finished.  Obviously, nothing was written, but just run this script
        again whenever you want to generate a template for a plugin.""")

if __name__ == '__main__':
    main()

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



#!/usr/bin/env python3

###
# Copyright (c) 2005, Jeremiah Fincher
# Copyright (c) 2009, James McCoy
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import os
import sys
import optparse
import subprocess

# XXX I wanted this for conf.version, but this will create directories. We
# really need to refactor conf so it either doesn't create directories, or
# so that static information (like the version) can be imported from
# somewhere else.
# from supybot import conf

VERBOSE = False

def readPid(filename):
    fd = open(filename)
    try:
        return int(fd.read().strip())
    finally:
        fd.close()

def isAlive(pid):
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        return False

def debug(s):
    if VERBOSE:
        if not s.endswith(os.linesep):
            s += os.linesep
        sys.stdout.write(s)

def main():
    parser = optparse.OptionParser(usage='Usage: %prog [options]')
    parser.add_option('', '--verbose', action='store_true',
                      help='Makes output verbose.')
    parser.add_option('', '--botdir',
                      help='Determines what directory the bot resides in and '
                      'should be started from.')
    parser.add_option('', '--pidfile',
                      help='Determines what file to look in for the pid of '
                      'the running bot.  This should be relative to the '
                      'given bot directory. Note that for this to actually '
                      'work, you have to make a matching entry in the '
                      'supybot.pidFile config in the supybot registry.')
    parser.add_option('', '--supybot', default='supybot',
                      help='Determines where the supybot executable is '
                      'located.  If not given, assumes that supybot is '
                      'in $PATH.')
    parser.add_option('', '--conffile',
                      help='Determines what configuration file should be '
                      'given to the supybot executable when (re)starting the '
                      'bot.')

    (options, args) = parser.parse_args()
    VERBOSE = options.verbose

    if args:
        parser.error('Extra arguments given.')
    if not options.botdir:
        parser.error('No botdir given.')
    if not options.pidfile:
        parser.error('No pidfile given.')
    if not options.conffile:
        parser.error('No conffile given.')

    os.chdir(options.botdir)
    open(options.pidfile, 'a').close() 

    pid = None
    try:
        pid = readPid(options.pidfile)
        debug('Found pidFile with proper pid contents of %s' % pid)
    except ValueError as e:
        foundBot = False

    if pid is not None:
        foundBot = isAlive(pid)
        if foundBot:
            debug('Pid %s is alive and belongs to us.' % pid)
        else:
            debug('Pid %s is not the bot.' % pid)

    if not foundBot:
        # First, we check if the pidfile is writable.  If not, supybot will just exit,
        # so we go ahead and refuse to start it.
        try:
            open(options.pidfile, 'r+')
        except EnvironmentError as e:
            debug('pidfile (%s) is not writable: %s' % (options.pidfile, e))
            sys.exit(-1)
        debug('Bot not found, starting.')
        cmdline = [options.supybot, '--daemon', options.conffile]
        inst = subprocess.Popen(cmdline, close_fds=True,
                                stderr=subprocess.STDOUT,
                                stdin=None, stdout=subprocess.PIPE)
        debug('Output from supybot: %r' % inst.stdout.read())
        ret = inst.wait()
        debug('Bot started, command line %r returned %s.' % (' '.join(cmdline),
                                                             ret))
        sys.exit(ret)
    else:
        sys.exit(0)

if __name__ == '__main__':
    main()

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



#!/usr/bin/env python3

import functools
import re

@functools.lru_cache
def makeSedRegex(delimiters=''):
    delimiters = delimiters or r"^\w\s"

    return re.compile(
        # This part matches an optional nick followed by ":" or ",", used to direct replacement
        # at a particular user.
        r"^(?:(?P<nick>.+?)[:,] )?"

        # Match and save the delimiter (any one symbol) as a named group
        fr"s(?P<delim>[{delimiters}])"

        # Match the pattern to replace, which can be any string up to the first instance of the
        # delimiter
        r"(?P<pattern>(?:(?!(?P=delim)).)*)(?P=delim)"

        # Ditto with the replacement
        r"(?P<replacement>(?:(?!(?P=delim)).)*)"

        # Optional final delimiter plus flags at the end
        r"(?:(?P=delim)(?P<flags>[a-z]*))?"
    )

if __name__ == '__main__':
    print("This is the full regex used by the plugin; paste it into your favourite regex tester "
          "for debugging:")
    print(makeSedRegex())



#!/usr/bin/env python3

###
# Copyright (c) 2002-2005, Jeremiah Fincher
# Copyright (c) 2011, James McCoy
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import os
import sys
import time
import shutil
import fnmatch
from tempfile import TemporaryDirectory
started = time.time()

import supybot
import logging
import traceback

# We need to do this before we import conf.
main_temp_dir = TemporaryDirectory()

os.makedirs(os.path.join(main_temp_dir.name, 'conf'))
os.makedirs(os.path.join(main_temp_dir.name, 'data'))
os.makedirs(os.path.join(main_temp_dir.name, 'logs'))

registryFilename = os.path.join(main_temp_dir.name, 'conf', 'test.conf')
with open(registryFilename, 'w') as fd:
    fd.write("""
supybot.directories.backup: /dev/null
supybot.directories.conf: {temp_conf}
supybot.directories.data: {temp_data}
supybot.directories.log: {temp_logs}
supybot.reply.whenNotCommand: True
supybot.log.stdout: False
supybot.log.stdout.level: ERROR
supybot.log.level: DEBUG
supybot.log.format: %(levelname)s %(message)s
supybot.log.plugins.individualLogfiles: False
supybot.protocols.irc.throttleTime: 0
supybot.reply.whenAddressedBy.chars: @
supybot.networks.test.server: should.not.need.this
supybot.networks.testnet1.server: should.not.need.this
supybot.networks.testnet2.server: should.not.need.this
supybot.networks.testnet3.server: should.not.need.this
supybot.nick: test
supybot.databases.users.allowUnregistration: True
""".format(
        temp_conf=os.path.join(main_temp_dir.name, 'conf'),
        temp_data=os.path.join(main_temp_dir.name, 'data'),
        temp_logs=os.path.join(main_temp_dir.name, 'logs')
    ))

import supybot.registry as registry
registry.open_registry(registryFilename)

import supybot.log as log
import supybot.conf as conf
conf.allowEval = True
conf.supybot.flush.setValue(False)

import re
import sys
import glob
import atexit
import os.path
import unittest

import supybot.utils as utils
import supybot.world as world
import supybot.callbacks as callbacks
world.startedAt = started

import logging
class TestLogFilter(logging.Filter):
    bads = [
        'No callbacks in',
        'Invalid channel database',
        'Exact error',
        'Invalid user dictionary',
        'because of noFlush',
        'Queuing NICK',
        'Queuing USER',
        'IgnoresDB.reload failed',
        'Starting log for',
        'Irc object for test dying',
        'Last Irc,',
        ]
    def filter(self, record):
        for bad in self.bads:
            if bad in record.msg:
                return False
        return True
log._logger.addFilter(TestLogFilter())

class path(str):
    """A class to represent platform-independent paths."""
    _r = re.compile(r'[\\/]')
    def __hash__(self):
        return reduce(lambda h, s: h ^ hash(s), self._r.split(self), 0)
    def __eq__(self, other):
        return self._r.split(self) == self._r.split(other)

def main():
    import glob
    import os.path
    import optparse
    import supybot.test as test
    import supybot.plugin as plugin

    parser = optparse.OptionParser(usage='Usage: %prog [options] [plugins]',
                                   version='Supybot %s' % conf.version)
    parser.add_option('-c', '--clean', action='store_true', default=False,
                      dest='clean', help='Cleans the various data/conf/logs'
                      'directories before running tests.')
    parser.add_option('--clean-after', action='store_true', default=False,
                      dest='clean_after', help='Cleans the various data/conf/logs'
                      'directories after running tests.')
    parser.add_option('-t', '--timeout', action='store', type='float',
                      dest='timeout',
                      help='Sets the timeout, in seconds, for tests to return '
                      'responses.')
    parser.add_option('-v', '--verbose', action='count', default=0,
                      help='Increase verbosity, logging extra information '
                           'about each test that runs.')
    parser.add_option('', '--fail-fast', action='store_true', default=False,
                      help='Stop at first failed test.')
    parser.add_option('', '--no-network', action='store_true', default=False,
                      dest='nonetwork', help='Causes the network-based tests '
                                             'not to run.')
    parser.add_option('', '--no-setuid', action='store_true', default=False,
                      dest='nosetuid', help='Causes the tests based on a '
                                             'setuid executable not to run.')
    parser.add_option('', '--trace', action='store_true', default=False,
                      help='Traces all calls made.  Unless you\'re really in '
                      'a pinch, you probably shouldn\'t do this; it results '
                      'in copious amounts of output.')
    parser.add_option('', '--plugins-dir',
                      action='append', dest='pluginsDirs', default=[],
                      help='Looks in in the given directory for plugins and '
                      'loads the tests from all of them.')
    parser.add_option('', '--exclude',
                      action='append', dest='excludePlugins', default=[],
                      help='List of plugins you do not want --plugins-dir '
                      'to include.')
    parser.add_option('', '--disable-multiprocessing', action='store_true',
                      dest='disableMultiprocessing',
                      help='Disables multiprocessing stuff.')
    (options, args) = parser.parse_args()

    world.disableMultiprocessing = options.disableMultiprocessing

    # This must go before checking for args, of course.
    if options.excludePlugins:
        exclusionPattern = re.compile('|'.join(
            fnmatch.translate(pat) for pat in options.excludePlugins
        ))
    else:
        exclusionPattern = None
    for pluginDir in options.pluginsDirs:
        for name in glob.glob(os.path.join(pluginDir, '*')):
            if (exclusionPattern is None or not exclusionPattern.match(name)) \
                    and os.path.isdir(name):
                args.append(name)

    if not args:
        parser.print_help()
        sys.exit(-1)

    if options.timeout:
        test.timeout = options.timeout

    if options.trace:
        traceFilename = conf.supybot.directories.log.dirize('trace.log')
        fd = open(traceFilename, 'w')
        sys.settrace(utils.gen.callTracer(fd))
        atexit.register(fd.close)
        atexit.register(lambda : sys.settrace(None))

    world.myVerbose = options.verbose

    if options.nonetwork:
        test.network = False
    if options.nosetuid:
        test.setuid = False

    log.testing = True
    world.testing = True

    args = [s.rstrip('\\/') for s in args]
    pluginDirs = set([os.path.dirname(s) or '.' for s in args])
    conf.supybot.directories.plugins.setValue(list(pluginDirs))
    pluginNames = set([os.path.basename(s) for s in args])
    load = unittest.defaultTestLoader.loadTestsFromModule
    for pluginName in pluginNames:
        if pluginName.endswith('.py'):
            pluginName = pluginName[:-3]
        try:
            pluginModule = plugin.loadPluginModule(pluginName)
        except (ImportError, callbacks.Error) as e:
            sys.stderr.write('Failed to load plugin %s:' % pluginName)
            traceback.print_exc()
            sys.stderr.write('(pluginDirs: %s)\n' %
                             conf.supybot.directories.plugins())
            continue
        if hasattr(pluginModule, 'test'):
            test.suites.append(load(pluginModule.test))

    suite = unittest.TestSuite(test.suites)
    if options.fail_fast:
        runner = unittest.TextTestRunner(verbosity=2,
                failfast=True)
    else:
        runner = unittest.TextTestRunner(verbosity=2)
    print('Testing began at %s (pid %s)' % (time.ctime(), os.getpid()))
    if options.clean:
        shutil.rmtree(conf.supybot.directories.log())
        log._mkDirs()
        shutil.rmtree(conf.supybot.directories.conf())
        shutil.rmtree(conf.supybot.directories.data())
    result = runner.run(suite)

    if hasattr(unittest, 'asserts'):
        print('Total asserts: %s' % unittest.asserts)

    if options.clean_after:
        log.setLevel(100)  # don't log anything anymore
        shutil.rmtree(conf.supybot.directories.log())
        shutil.rmtree(conf.supybot.directories.conf())
        shutil.rmtree(conf.supybot.directories.data())

    if result.wasSuccessful():
        sys.exit(0)
    else:
        # Deactivate autocleaning for the temporary directiories to allow inspection.
        main_temp_dir._finalizer.detach()
        print(f"Temporary directory path: {main_temp_dir.name}")
        sys.exit(1)


if __name__ == '__main__':
    main()

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2002-2004, Jeremiah Fincher
# Copyright (c) 2009, James McCoy
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import functools
from unittest.mock import patch
import socket
import sys

import feedparser

from supybot.test import *
import supybot.conf as conf
import supybot.utils.minisix as minisix

xkcd_old = """<?xml version="1.0" encoding="utf-8"?>
<rss version="2.0"><channel><title>xkcd.com</title><link>http://xkcd.com/</link><description>xkcd.com: A webcomic of romance and math humor.</description><language>en</language><item><title>Snake Facts</title><link>http://xkcd.com/1398/</link><description>&lt;img src="http://imgs.xkcd.com/comics/snake_facts.png" title="Biologically speaking, what we call a 'snake' is actually a human digestive tract which has escaped from its host." alt="Biologically speaking, what we call a 'snake' is actually a human digestive tract which has escaped from its host." /&gt;</description><pubDate>Wed, 23 Jul 2014 04:00:00 -0000</pubDate><guid>http://xkcd.com/1398/</guid></item></channel></rss>
"""

xkcd_new = """<?xml version="1.0" encoding="utf-8"?>
<rss version="2.0"><channel><title>xkcd.com</title><link>http://xkcd.com/</link><description>xkcd.com: A webcomic of romance and math humor.</description><language>en</language><item><title>Telescopes: Refractor vs Reflector</title><link>http://xkcd.com/1791/</link><description>&lt;img src="http://imgs.xkcd.com/comics/telescopes_refractor_vs_reflector.png" title="On the other hand, the refractor's limited light-gathering means it's unable to make out shadow people or the dark god Chernabog." alt="On the other hand, the refractor's limited light-gathering means it's unable to make out shadow people or the dark god Chernabog." /&gt;</description><pubDate>Fri, 27 Jan 2017 05:00:00 -0000</pubDate><guid>http://xkcd.com/1791/</guid></item><item><title>Chaos</title><link>http://xkcd.com/1399/</link><description>&lt;img src="http://imgs.xkcd.com/comics/chaos.png" title="Although the oral exam for the doctorate was just 'can you do that weird laugh?'" alt="Although the oral exam for the doctorate was just 'can you do that weird laugh?'" /&gt;</description><pubDate>Fri, 25 Jul 2014 04:00:00 -0000</pubDate><guid>http://xkcd.com/1399/</guid></item><item><title>Snake Facts</title><link>http://xkcd.com/1398/</link><description>&lt;img src="http://imgs.xkcd.com/comics/snake_facts.png" title="Biologically speaking, what we call a 'snake' is actually a human digestive tract which has escaped from its host." alt="Biologically speaking, what we call a 'snake' is actually a human digestive tract which has escaped from its host." /&gt;</description><pubDate>Wed, 23 Jul 2014 04:00:00 -0000</pubDate><guid>http://xkcd.com/1398/</guid></item></channel></rss>
"""

not_well_formed = """<?xml version="1.0" encoding="utf-8"?>
<rss version="2.0">
<channel>
	<title>this is missing a close tag
	<link>http://example.com/</link>
	<description>this dummy feed has no elements</description>
	<language>en</language>
</channel>
</rss>
"""

class MockResponse:
    headers = {}
    url = ''
    def read(self):
        return self._data.encode()

    def close(self):
        pass

    def geturl(self):
        return url

def mock_urllib(f):
    mock = MockResponse()

    @functools.wraps(f)
    def newf(self):
        with patch("urllib.request.OpenerDirector.open", return_value=mock):
            f(self, mock)

    return newf


url = 'http://www.advogato.org/rss/articles.xml'
class RSSTestCase(ChannelPluginTestCase):
    plugins = ('RSS', 'Plugin')

    timeout = 1

    def testRssAddBadName(self):
        self.assertError('rss add "foo bar" %s' % url)

    def testCantAddFeedNamedRss(self):
        self.assertError('rss add rss %s' % url)

    def testCantRemoveMethodThatIsntFeed(self):
        self.assertError('rss remove rss')

    def testCantAddDuplicatedFeed(self):
        self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
        try:
            self.assertError('rss add xkcddup http://xkcd.com/rss.xml')
        finally:
            self.assertNotError('rss remove xkcd')

    @mock_urllib
    def testRemoveAliasedFeed(self, mock):
        mock._data = xkcd_old
        try:
            self.assertNotError('rss announce add http://xkcd.com/rss.xml')

            # Clear the queue or it's going to mess up the finally block
            self.assertRegexp(' ', 'Snake Facts')
            self.assertNoResponse(' ')

            self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
        finally:
            self.assertNotError('rss announce remove http://xkcd.com/rss.xml')
            self.assertNotError('rss remove xkcd')
        self.assertEqual(self.irc.getCallback('RSS').feed_names, {})
        self.assertTrue(self.irc.getCallback('RSS').get_feed('http://xkcd.com/rss.xml'))

    @mock_urllib
    def testChangeUrl(self, mock):
        try:
            self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
            self.assertNotError('rss remove xkcd')
            self.assertNotError('rss add xkcd https://xkcd.com/rss.xml')
            self.assertRegexp('help xkcd', 'https://')
        finally:
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testChangeName(self, mock):
        try:
            self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
            self.assertNotError('rss remove xkcd')
            self.assertNotError('rss add xkcd2 http://xkcd.com/rss.xml')
            self.assertRegexp('help xkcd2', 'http://xkcd.com')
        finally:
            self._feedMsg('rss remove xkcd')
            self._feedMsg('rss remove xkcd2')

    @mock_urllib
    def testInitialAnnounceNewest(self, mock):
        mock._data = xkcd_new
        timeFastForward(1.1)
        try:
            with conf.supybot.plugins.RSS.sortFeedItems.context('newestFirst'):
                with conf.supybot.plugins.RSS.initialAnnounceHeadlines.context(1):
                    self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
                    self.assertNotError('rss announce add xkcd')
                    self.assertRegexp(' ', 'Telescopes')
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testInitialAnnounceOldest(self, mock):
        mock._data = xkcd_new
        timeFastForward(1.1)
        try:
            with conf.supybot.plugins.RSS.initialAnnounceHeadlines.context(1):
                with conf.supybot.plugins.RSS.sortFeedItems.context('oldestFirst'):
                    self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
                    self.assertNotError('rss announce add xkcd')
                    self.assertRegexp(' ', 'Telescopes')
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testNoInitialAnnounce(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            with conf.supybot.plugins.RSS.initialAnnounceHeadlines.context(0):
                self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
                self.assertNotError('rss announce add xkcd')
                self.assertNoResponse(' ', timeout=0.1)
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testAnnounce(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            self.assertError('rss announce add xkcd')
            self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
            self.assertNotError('rss announce add xkcd')
            self.assertNotError(' ')
            with conf.supybot.plugins.RSS.sortFeedItems.context('oldestFirst'):
                with conf.supybot.plugins.RSS.waitPeriod.context(1):
                    timeFastForward(1.1)
                    self.assertNoResponse(' ', timeout=0.1)
                    mock._data = xkcd_new
                    self.assertNoResponse(' ', timeout=0.1)
                    timeFastForward(1.1)
                    self.assertRegexp(' ', 'Chaos')
                    self.assertRegexp(' ', 'Telescopes')
                    self.assertNoResponse(' ')
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testMaxAnnounces(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            self.assertError('rss announce add xkcd')
            self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
            self.assertNotError('rss announce add xkcd')
            self.assertNotError(' ')
            with conf.supybot.plugins.RSS.sortFeedItems.context('oldestFirst'):
                with conf.supybot.plugins.RSS.waitPeriod.context(1):
                    with conf.supybot.plugins.RSS.maximumAnnounceHeadlines.context(1):
                        timeFastForward(1.1)
                        self.assertNoResponse(' ', timeout=0.1)
                        mock._data = xkcd_new
                        log.debug('set return value to: %r', xkcd_new)
                        self.assertNoResponse(' ', timeout=0.1)
                        timeFastForward(1.1)
                        self.assertRegexp(' ', 'Telescopes')
                        self.assertNoResponse(' ')
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testAnnounceAnonymous(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            self.assertNotError('rss announce add http://xkcd.com/rss.xml')
            self.assertNotError(' ')
            with conf.supybot.plugins.RSS.waitPeriod.context(1):
                timeFastForward(1.1)
                self.assertNoResponse(' ', timeout=0.1)
                mock._data = xkcd_new
                self.assertNoResponse(' ', timeout=0.1)
                timeFastForward(1.1)
                self.assertRegexp(' ', 'Telescopes')
                self.assertRegexp(' ', 'Chaos')
                self.assertNoResponse(' ', timeout=0.1)
            self.assertResponse('announce list', 'http://xkcd.com/rss.xml')
        finally:
            self._feedMsg('rss announce remove http://xkcd.com/rss.xml')
            self._feedMsg('rss remove http://xkcd.com/rss.xml')

    @mock_urllib
    def testAnnounceReload(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            with conf.supybot.plugins.RSS.waitPeriod.context(1):
                self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
                self.assertNotError('rss announce add xkcd')
                self.assertNotError(' ', timeout=0.1)
                self.assertNotError('reload RSS')
                self.assertNoResponse(' ', timeout=0.1)
                timeFastForward(1.1)
                self.assertNoResponse(' ', timeout=0.1)
            self.assertResponse('announce list', 'xkcd')
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testReload(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            with conf.supybot.plugins.RSS.waitPeriod.context(1):
                self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
                self.assertNotError('rss announce add xkcd')
                self.assertRegexp(' ', 'Snake Facts')
                mock._data = xkcd_new
                self.assertNotError('reload RSS')
                self.assertRegexp(' ', 'Telescopes')
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testReloadNoDelay(self, mock):
        # https://github.com/progval/Limnoria/issues/922
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            with conf.supybot.plugins.RSS.waitPeriod.context(1):
                self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
                self.assertRegexp('xkcd', 'Snake Facts')
                self.assertNotError('reload RSS')
                self.assertRegexp('xkcd', 'Snake Facts')
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testReannounce(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            self.assertError('rss announce add xkcd')
            self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
            self.assertNotError('rss announce add xkcd')
            self.assertRegexp(' ', 'Snake Facts')
            with conf.supybot.plugins.RSS.waitPeriod.context(1):
                with conf.supybot.plugins.RSS.initialAnnounceHeadlines.context(1):
                    with conf.supybot.plugins.RSS.sortFeedItems.context('oldestFirst'):
                        timeFastForward(1.1)
                        self.assertNoResponse(' ', timeout=0.1)
                        self._feedMsg('rss announce remove xkcd')
                        mock._data = xkcd_new
                        timeFastForward(1.1)
                        self.assertNoResponse(' ', timeout=0.1)
                        self.assertNotError('rss announce add xkcd')
                        timeFastForward(1.1)
                        self.assertRegexp(' ', 'Chaos')
                        self.assertRegexp(' ', 'Telescopes')
                        self.assertNoResponse(' ')
        finally:
            self._feedMsg('rss announce remove xkcd')
            self._feedMsg('rss remove xkcd')

    @mock_urllib
    def testFeedSpecificFormat(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            self.assertNotError('rss add xkcd http://xkcd.com/rss.xml')
            self.assertNotError('rss add xkcdsec https://xkcd.com/rss.xml')
            self.assertNotError('config plugins.RSS.feeds.xkcd.format foo')
            self.assertRegexp('config plugins.RSS.feeds.xkcd.format', 'foo')
            self.assertRegexp('xkcd', 'foo')
            self.assertNotRegexp('xkcdsec', 'foo')
        finally:
            self._feedMsg('rss remove xkcd')
            self._feedMsg('rss remove xkcdsec')

    @mock_urllib
    def testFeedSpecificWaitPeriod(self, mock):
        mock._data = xkcd_old
        timeFastForward(1.1)
        try:
            self.assertNotError('rss add xkcd1 http://xkcd.com/rss.xml')
            self.assertNotError('rss announce add xkcd1')
            self.assertNotError('rss add xkcd2 http://xkcd.com/rss.xml&foo')
            self.assertNotError('rss announce add xkcd2')
            self.assertNotError(' ')
            self.assertNotError(' ')
            with conf.supybot.plugins.RSS.sortFeedItems.context('oldestFirst'):
                with conf.supybot.plugins.RSS.feeds.xkcd1.waitPeriod.context(1):
                    timeFastForward(1.1)
                    self.assertNoResponse(' ', timeout=0.1)
                    mock._data = xkcd_new
                    self.assertNoResponse(' ', timeout=0.1)
                    timeFastForward(1.1)
                    self.assertRegexp(' ', 'xkcd1.*Chaos')
                    self.assertRegexp(' ', 'xkcd1.*Telescopes')
                    self.assertNoResponse(' ')
                    timeFastForward(1.1)
                    self.assertNoResponse(' ', timeout=0.1)
        finally:
            self._feedMsg('rss announce remove xkcd1')
            self._feedMsg('rss remove xkcd1')
            self._feedMsg('rss announce remove xkcd2')
            self._feedMsg('rss remove xkcd2')

    @mock_urllib
    def testDescription(self, mock):
        timeFastForward(1.1)
        with conf.supybot.plugins.RSS.format.context('$description'):
            mock._data = xkcd_new
            self.assertRegexp('rss http://xkcd.com/rss.xml',
                    'On the other hand, the refractor\'s')

    @mock_urllib
    def testAtomContentHtmlOnly(self, mock):
        timeFastForward(1.1)
        mock._data = """
<?xml version="1.0" encoding="UTF-8"?>
<feed xmlns="http://www.w3.org/2005/Atom" xmlns:media="http://search.yahoo.com/mrss/" xml:lang="en-US">
  <title>Recent Commits to anope:2.0</title>
  <updated>2023-10-04T16:14:39Z</updated>
  <entry>
    <title>title with &lt;pre&gt;HTML&lt;/pre&gt;</title>
    <updated>2023-10-04T16:14:39Z</updated>
    <content type="html">
      content with &lt;pre&gt;HTML&lt;/pre&gt;
    </content>
  </entry>
</feed>"""
        with conf.supybot.plugins.RSS.format.context('$content'):
            self.assertRegexp('rss https://example.org',
                    'content with HTML')
        with conf.supybot.plugins.RSS.format.context('$description'):
            self.assertRegexp('rss https://example.org',
                    'content with HTML')

    @mock_urllib
    def testAtomContentXhtmlOnly(self, mock):
        timeFastForward(1.1)
        mock._data = """
<?xml version="1.0" encoding="UTF-8"?>
<feed xmlns="http://www.w3.org/2005/Atom" xmlns:media="http://search.yahoo.com/mrss/" xml:lang="en-US">
  <title>Recent Commits to anope:2.0</title>
  <updated>2023-10-04T16:14:39Z</updated>
  <entry>
    <title>title with &lt;pre&gt;HTML&lt;/pre&gt;</title>
    <updated>2023-10-04T16:14:39Z</updated>
    <content type="xhtml">
      <div xmlns="http://www.w3.org/1999/xhtml">
        content with <pre>XHTML</pre>
      </div>
    </content>
  </entry>
</feed>"""
        with conf.supybot.plugins.RSS.format.context('$content'):
            self.assertRegexp('rss https://example.org',
                    'content with XHTML')
        with conf.supybot.plugins.RSS.format.context('$description'):
            self.assertRegexp('rss https://example.org',
                    'content with XHTML')

    @mock_urllib
    def testAtomContentHtmlAndPlaintext(self, mock):
        timeFastForward(1.1)
        mock._data = """
<?xml version="1.0" encoding="UTF-8"?>
<feed xmlns="http://www.w3.org/2005/Atom" xmlns:media="http://search.yahoo.com/mrss/" xml:lang="en-US">
  <title>Recent Commits to anope:2.0</title>
  <updated>2023-10-04T16:14:39Z</updated>
  <entry>
    <title>title with &lt;pre&gt;HTML&lt;/pre&gt;</title>
    <updated>2023-10-04T16:14:39Z</updated>
    <!-- Atom spec says multiple contents is invalid, feedparser says it's not.
         I like having the option, so let's make sure we support it. -->
    <content type="html">
      content with &lt;pre&gt;HTML&lt;/pre&gt;
    </content>
    <content type="text">
      content with plaintext
    </content>
  </entry>
</feed>"""
        with conf.supybot.plugins.RSS.format.context('$content'):
            self.assertRegexp('rss https://example.org',
                    'content with plaintext')
        with conf.supybot.plugins.RSS.format.context('$description'):
            self.assertRegexp('rss https://example.org',
                    'content with plaintext')

    @mock_urllib
    def testAtomContentPlaintextAndHtml(self, mock):
        timeFastForward(1.1)
        mock._data = """
<?xml version="1.0" encoding="UTF-8"?>
<feed xmlns="http://www.w3.org/2005/Atom" xmlns:media="http://search.yahoo.com/mrss/" xml:lang="en-US">
  <title>Recent Commits to anope:2.0</title>
  <updated>2023-10-04T16:14:39Z</updated>
  <entry>
    <title>title with &lt;pre&gt;HTML&lt;/pre&gt;</title>
    <updated>2023-10-04T16:14:39Z</updated>
    <!-- Atom spec says multiple contents is invalid, feedparser says it's not.
         I like having the option, so let's make sure we support it. -->
    <content type="text">
      content with plaintext
    </content>
    <content type="html">
      content with &lt;pre&gt;HTML&lt;/pre&gt;
    </content>
  </entry>
</feed>"""
        with conf.supybot.plugins.RSS.format.context('$content'):
            self.assertRegexp('rss https://example.org',
                    'content with plaintext')
        with conf.supybot.plugins.RSS.format.context('$description'):
            self.assertRegexp('rss https://example.org',
                    'content with plaintext')

    @mock_urllib
    def testRssDescriptionHtml(self, mock):
        timeFastForward(1.1)
        mock._data = """
<?xml version="1.0" encoding="utf-8"?>
<rss xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:atom="http://www.w3.org/2005/Atom" xmlns:content="http://purl.org/rss/1.0/modules/content/" xmlns:foaf="http://xmlns.com/foaf/0.1/" xmlns:og="http://ogp.me/ns#" xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#" xmlns:schema="http://schema.org/" xmlns:sioc="http://rdfs.org/sioc/ns#" xmlns:sioct="http://rdfs.org/sioc/types#" xmlns:skos="http://www.w3.org/2004/02/skos/core#" xmlns:xsd="http://www.w3.org/2001/XMLSchema#" version="2.0">
  <channel>
    <title>feed title</title>
    <description/>
    <language>en</language>
    <item>
    <title>title with &lt;pre&gt;HTML&lt;/pre&gt;</title>
    <description>description with &lt;pre&gt;HTML&lt;/pre&gt;</description>
    </item>
  </channel>
</feed>"""
        with conf.supybot.plugins.RSS.format.context('$description'):
            self.assertRegexp('rss https://example.org',
                    'description with HTML')

    @mock_urllib
    def testFeedAttribute(self, mock):
        timeFastForward(1.1)
        with conf.supybot.plugins.RSS.format.context('$feed_title: $title'):
            mock._data = xkcd_new
            self.assertRegexp('rss http://xkcd.com/rss.xml',
                              r'xkcd\.com: Telescopes')

    @mock_urllib
    def testBadlyFormedFeedWithNoItems(self, mock):
        # This combination will cause the RSS command to show the last parser
        # error.
        timeFastForward(1.1)
        mock._data = not_well_formed
        self.assertRegexp('rss http://example.com/',
                          'Parser error: .*mismatch')

    def testSocketError(self):
        class MockResponse:
            headers = {}
            url = ''
            def read(self):
                raise socket.error("oh no")

            def close(self):
                pass
        mock = MockResponse()
        with patch("urllib.request.OpenerDirector.open", return_value=mock):
            timeFastForward(1.1)
            self.assertRegexp('rss http://example.com/',
                              'Parser error: .*oh no')

    if network:
        timeout = 5  # Note this applies also to the above tests

        def testRssinfo(self):
            timeFastForward(1.1)
            self.assertNotError('rss info %s' % url)
            self.assertNotError('rss add advogato %s' % url)
            self.assertNotError('rss info advogato')
            self.assertNotError('rss info AdVogATo')
            self.assertNotError('rss remove advogato')

        def testRssinfoDoesTimeProperly(self):
            timeFastForward(1.1)
            self.assertNotRegexp('rss info http://slashdot.org/slashdot.rss',
                                 '-1 years')

        def testAnnounceAdd(self):
            timeFastForward(1.1)
            self.assertNotError('rss add advogato %s' % url)
            self.assertNotError('rss announce add advogato')
            self.assertRegexp('rss announce channels advogato', 'advogato is announced to.*%s%s' % (self.irc.network, self.channel))

            self.assertNotRegexp('rss announce', r'ValueError')

            self.assertNotError('rss announce remove advogato')
            self.assertRegexp('rss announce channels advogato', 'advogato is not announced to any channels')

            self.assertNotError('rss remove advogato')
            self.assertRegexp('rss announce channels advogato', 'Unknown feed')

        def testRss(self):
            timeFastForward(1.1)
            self.assertNotError('rss %s' % url)
            m = self.assertNotError('rss %s 2' % url)
            self.assertEqual(m.args[1].count(' | '), 1)

        def testRssRandom(self):
            timeFastForward(1.1)
            self.assertNotError('rss --random %s' % url)
            m = self.assertNotError('rss --random %s 2' % url)
            self.assertEqual(m.args[1].count(' | '), 1)

        def testRssAdd(self):
            timeFastForward(1.1)
            self.assertNotError('rss add advogato %s' % url)
            self.assertNotError('advogato')
            self.assertNotError('rss advogato')
            self.assertNotError('rss remove advogato')
            self.assertNotRegexp('list RSS', 'advogato')
            self.assertError('advogato')
            self.assertError('rss advogato')

        def testNonAsciiFeeds(self):
            timeFastForward(1.1)
            self.assertNotError('rss http://www.heise.de/newsticker/heise.rdf')
            self.assertNotError('rss info http://br-linux.org/main/index.xml')



# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



#!/usr/bin/env python3

###
# Copyright (c) 2003-2004, Jeremiah Fincher
# Copyright (c) 2009, James McCoy
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

from __future__ import print_function

import os
import sys

def error(s):
    sys.stderr.write(s)
    if not s.endswith(os.linesep):
        sys.stderr.write(os.linesep)
    sys.exit(-1)

import supybot

import re
import time
import pydoc
import pprint
import socket
import logging
import optparse

try:
    import supybot.i18n as i18n
except ImportError:
    sys.stderr.write("""Error:
    You are running a mix of Limnoria and stock Supybot code. Although you run
    one of Limnoria\'s executables, Python tries to load stock
    Supybot\'s library. To fix this issue, uninstall Supybot
    ("%s -m pip uninstall supybot" should do the job)
    and install Limnoria again.
    For your information, Supybot's libraries are installed here:
    %s\n""" %
    (sys.executable, '\n    '.join(supybot.__path__)))
    exit(-1)

import supybot.ansi as ansi
import supybot.utils as utils
import supybot.ircutils as ircutils
import supybot.registry as registry
# supybot.plugin, supybot.log, and supybot.conf will be imported later,
# because we need to set a language before loading the conf

import supybot.questions as questions
from supybot.questions import output, yn, anything, something, expect, getpass

def getPlugins(pluginDirs):
    plugins = set([])
    join = os.path.join
    for pluginDir in pluginDirs:
        try:
            for filename in os.listdir(pluginDir):
                fname = join(pluginDir, filename)
                if (filename.endswith('.py') or os.path.isdir(fname)) \
                   and filename[0].isupper():
                    plugins.add(os.path.splitext(filename)[0])
        except OSError:
            continue
    plugins.discard('Owner')
    plugins = list(plugins)
    plugins.sort()
    return plugins

def loadPlugin(name):
    import supybot.plugin as plugin
    try:
        module = plugin.loadPluginModule(name)
        if hasattr(module, 'Class'):
            return module
        else:
            output("""That plugin loaded fine, but didn't seem to be a real
            Supybot plugin; there was no Class variable to tell us what class
            to load when we load the plugin.  We'll skip over it for now, but
            you can always add it later.""")
            return None
    except Exception as e:
        output("""We encountered a bit of trouble trying to load plugin %r.
        Python told us %r.  We'll skip over it for now, you can always add it
        later.""" % (name, utils.gen.exnToString(e)))
        return None

def describePlugin(module, showUsage):
    if module.__doc__:
        output(module.__doc__, unformatted=False)
    elif hasattr(module.Class, '__doc__'):
        output(module.Class.__doc__, unformatted=False)
    else:
        output("""Unfortunately, this plugin doesn't seem to have any
        documentation.  Sorry about that.""")
    if showUsage:
        if hasattr(module, 'example'):
            if yn('This plugin has a usage example.  '
                  'Would you like to see it?', default=False):
                pydoc.pager(module.example)
        else:
            output("""This plugin has no usage example.""")

def clearLoadedPlugins(plugins, pluginRegistry):
    for plugin in plugins:
        try:
            pluginKey = pluginRegistry.get(plugin)
            if pluginKey():
                plugins.remove(plugin)
        except registry.NonExistentRegistryEntry:
            continue

_windowsVarRe = re.compile(r'%(\w+)%')
def getDirectoryName(default, basedir=os.curdir, prompt=True):
    done = False
    while not done:
        if prompt:
            dir = something('What directory do you want to use?',
                           default=os.path.join(basedir, default))
        else:
            dir = os.path.join(basedir, default)
        orig_dir = dir
        dir = os.path.expanduser(dir)
        dir = _windowsVarRe.sub(r'$\1', dir)
        dir = os.path.expandvars(dir)
        dir = os.path.abspath(dir)
        try:
            os.makedirs(dir)
            done = True
        except OSError as e:
            # 17 is File exists for Linux (and likely other POSIX systems)
            # 183 is the same for Windows
            if e.args[0] == 17 or (os.name == 'nt' and e.args[0] == 183):
                done = True
            else:
                output("""Sorry, I couldn't make that directory for some
                reason.  The Operating System told me %s.  You're going to
                have to pick someplace else.""" % e)
                prompt = True
    return (dir, os.path.dirname(orig_dir))


def _main():
    import supybot.version as version
    parser = optparse.OptionParser(usage='Usage: %prog [options]',
                                   version='Supybot %s' % version.version)
    parser.add_option('', '--allow-root', action='store_true',
                      dest='allowRoot',
                      help='Determines whether the wizard will be allowed to '
                           'as root. This should not be used except in special '
                           'circumstances, such as running inside a containerized '
                           'environment.')
    parser.add_option('', '--allow-home', action='store_true',
                      dest='allowHome',
                      help='Determines whether the wizard will be allowed to '
                           'run directly in the HOME directory. '
                           'You should not do this unless you want it to '
                           'create multiple files in your HOME directory.')
    parser.add_option('', '--allow-bin', action='store_true',
                      dest='allowBin',
                      help='Determines whether the wizard will be allowed to '
                           'run directly in a directory for binaries (eg. '
                           '/usr/bin, /usr/local/bin, or ~/.local/bin). '
                           'You should not do this unless you want it to '
                           'create multiple non-binary files in directory '
                           'that should only contain binaries.')
    parser.add_option('', '--no-network', action='store_false',
                      dest='network',
                      help='Determines whether the wizard will be allowed to '
                           'run without a network connection.')
    (options, args) = parser.parse_args()
    if os.name == 'posix':
        if (os.getuid() == 0 or os.geteuid() == 0) and not options.allowRoot:
            error('Running as root is not supported by default (see --allow-root).')
    if os.name == 'posix':
        if (os.getcwd() == os.path.expanduser('~')) and not options.allowHome:
            error('Please, don\'t run this in your HOME directory.')
        if (os.path.split(os.getcwd())[-1] == 'bin') and not options.allowBin:
            error('Please, don\'t run this in a "bin" directory.')
    if os.path.isfile(os.path.join('scripts', 'supybot-wizard')) or \
            os.path.isfile(os.path.join('..', 'scripts', 'supybot-wizard')):
        print('')
        print('+------------------------------------------------------------+')
        print('| +--------------------------------------------------------+ |')
        print('| | Warning: It looks like you are running the wizard from | |')
        print('| | the Supybot source directory. This is not recommended. | |')
        print('| | Please press Ctrl-C and change to another directory.   | |')
        print('| +--------------------------------------------------------+ |')
        print('+------------------------------------------------------------+')
        print('')

    if args:
        parser.error('This program takes no non-option arguments.')
    output("""This is a wizard to help you start running supybot.  What it
    will do is create the necessary config files based on the options you
    select here.  So hold on tight and be ready to be interrogated :)""")


    output("""First of all, we can bold the questions you're asked so you can
    easily distinguish the mostly useless blather (like this) from the
    questions that you actually have to answer.""")
    if yn('Would you like to try this bolding?', default=True):
        questions.useBold = True
        if not yn('Do you see this in bold?'):
            output("""Sorry, it looks like your terminal isn't ANSI compliant.
            Try again some other day, on some other terminal :)""")
            questions.useBold = False
        else:
            output("""Great!""")

    ###
    # Preliminary questions.
    ###
    output("""We've got some preliminary things to get out of the way before
    we can really start asking you questions that directly relate to what your
    bot is going to be like.""")

    # Advanced?
    output("""We want to know if you consider yourself an advanced Supybot
    user because some questions are just utterly boring and useless for new
    users.  Others might not make sense unless you've used Supybot for some
    time.""")
    advanced = yn('Are you an advanced Supybot user?', default=False)

    # Language?
    output("""This version of Supybot (known as Limnoria) includes another
    language. This can be changed at any time. You need to answer with a short
    id for the language, such as 'en', 'fr', 'it' (without the quotes). If
    you want to use English, just press enter.""")
    language = expect('What language do you want to use?',
                      i18n.SUPPORTED_LANGUAGES, default='en')

    class Empty:
        """This is a hack to allow the i18n to get the current language, before
        loading the conf module, before the conf module needs i18n to set the
        default strings."""
        def __call__(self):
            return self.value
    fakeConf = Empty()
    fakeConf.supybot = Empty()
    fakeConf.supybot.language = Empty()
    fakeConf.supybot.language.value = language
    i18n.conf = fakeConf
    i18n.currentLocale = language
    i18n.reloadLocales()
    import supybot.conf as conf
    i18n.import_conf() # It imports the real conf module

    ### Directories.
    # We set these variables in cache because otherwise conf and log will
    # create directories for the default values, which might not be what the
    # user wants.
    if advanced:
        output("""Now we've got to ask you some questions about where some of
        your directories are (or, perhaps, will be :)).  If you're running this
        wizard from the directory you'll actually be starting your bot from and
        don't mind creating some directories in the current directory, then
        just don't give answers to these questions and we'll create the
        directories we need right here in this directory.""")

        # conf.supybot.directories.log
        output("""Your bot will need to put its logs somewhere.  Do you have
        any specific place you'd like them?  If not, just press enter and we'll
        make a directory named "logs" right here.""")
        (logDir, basedir) = getDirectoryName('logs')
        conf.supybot.directories.log.setValue(logDir)
        import supybot.log as log
        log._stdoutHandler.setLevel(100) # *Nothing* gets through this!

        # conf.supybot.directories.data
        output("""Your bot will need to put various data somewhere.  Things
        like databases, downloaded files, etc.  Do you have any specific place
        you'd like the bot to put these things?  If not, just press enter and
        we'll make a directory named "data" right here.""")
        (dataDir, basedir) = getDirectoryName('data', basedir=basedir)
        conf.supybot.directories.data.setValue(dataDir)

        # conf.supybot.directories.conf
        output("""Your bot must know where to find its configuration files.
        It'll probably only make one or two, but it's gotta have some place to
        put them.  Where should that place be?  If you don't care, just press
        enter and we'll make a directory right here named "conf" where it'll
        store its stuff. """)
        (confDir, basedir) = getDirectoryName('conf', basedir=basedir)
        conf.supybot.directories.conf.setValue(confDir)

        # conf.supybot.directories.backup
        output("""Your bot must know where to place backups of its conf and
        data files.  Where should that place be?  If you don't care, just press
        enter and we'll make a directory right here named "backup" where it'll
        store its stuff.""")
        (backupDir, basedir) = getDirectoryName('backup', basedir=basedir)
        conf.supybot.directories.backup.setValue(backupDir)

        # conf.supybot.directories.data.tmp
        output("""Your bot needs a directory to put temporary files (used
        mainly to atomically update its configuration files).""")
        (tmpDir, basedir) = getDirectoryName('tmp', basedir=basedir)
        conf.supybot.directories.data.tmp.setValue(tmpDir)

        # conf.supybot.directories.data.web
        output("""Your bot needs a directory to put files related to the web
        server (templates, CSS).""")
        (webDir, basedir) = getDirectoryName('web', basedir=basedir)
        conf.supybot.directories.data.web.setValue(webDir)


        # imports callbacks, which imports ircdb, which requires
        # directories.conf
        import supybot.plugin as plugin

        # pluginDirs
        output("""Your bot will also need to know where to find its plugins at.
        Of course, it already knows where the plugins that it came with are,
        but your own personal plugins that you write for will probably be
        somewhere else.""")
        pluginDirs = conf.supybot.directories.plugins()
        output("""Currently, the bot knows about the following directories:""")
        output(format('%L', pluginDirs + [plugin._pluginsDir]))
        while yn('Would you like to add another plugin directory?  '
                 'Adding a local plugin directory is good style.',
                 default=True):
            (pluginDir, _) = getDirectoryName('plugins', basedir=basedir)
            if pluginDir not in pluginDirs:
                pluginDirs.append(pluginDir)
        conf.supybot.directories.plugins.setValue(pluginDirs)
    else:
        output("""Your bot needs to create some directories in order to store
        the various log, config, and data files.""")
        basedir = something("""Where would you like to create these
                            directories?""", default=os.curdir)

        # conf.supybot.directories.log
        (logDir, basedir) = getDirectoryName('logs',
                                             basedir=basedir, prompt=False)
        conf.supybot.directories.log.setValue(logDir)

        # conf.supybot.directories.data
        (dataDir, basedir) = getDirectoryName('data',
                                              basedir=basedir, prompt=False)
        conf.supybot.directories.data.setValue(dataDir)
        (tmpDir, basedir) = getDirectoryName('tmp',
                                             basedir=basedir, prompt=False)
        conf.supybot.directories.data.tmp.setValue(tmpDir)
        (webDir, basedir) = getDirectoryName('web',
                                             basedir=basedir, prompt=False)
        conf.supybot.directories.data.web.setValue(webDir)

        # conf.supybot.directories.conf
        (confDir, basedir) = getDirectoryName('conf',
                                              basedir=basedir, prompt=False)
        conf.supybot.directories.conf.setValue(confDir)

        # conf.supybot.directories.backup
        (backupDir, basedir) = getDirectoryName('backup',
                                                basedir=basedir, prompt=False)
        conf.supybot.directories.backup.setValue(backupDir)

        # pluginDirs
        pluginDirs = conf.supybot.directories.plugins()
        (pluginDir, _) = getDirectoryName('plugins',
                                          basedir=basedir, prompt=False)
        if pluginDir not in pluginDirs:
            pluginDirs.append(pluginDir)
        conf.supybot.directories.plugins.setValue(pluginDirs)

    import supybot.log as log
    log._stdoutHandler.setLevel(100) # *Nothing* gets through this!
    import supybot.plugin as plugin

    output("Good!  We're done with the directory stuff.")

    ###
    # Bot stuff
    ###
    output("""Now we're going to ask you things that actually relate to the
    bot you'll be running.""")

    network = None
    while not network:
        output("""First, we need to know the name of the network you'd like to
        connect to.  Not the server host, mind you, but the name of the
        network.  If you plan to connect to irc.libera.chat, for instance,
        you should answer this question with 'libera' or 'liberachat'
        (without the quotes).
        """)
        network = something('What IRC network will you be connecting to?')
        if '.' in network:
            output("""There shouldn't be a '.' in the network name.  Remember,
            this is the network name, not the actual server you plan to connect
            to.""")
            network = None
        elif not registry.isValidRegistryName(network):
            output("""That's not a valid name for one reason or another.  Please
            pick a simpler name, one more likely to be valid.""")
            network = None

    conf.supybot.networks.setValue([network])
    network = conf.registerNetwork(network)

    defaultServer = None
    server = None
    ip = None
    while not ip:
        serverString = something('What server would you like to connect to?',
                                 default=defaultServer)
        if options.network:
            try:
                output("""Looking up %s...""" % serverString)
                ip = socket.gethostbyname(serverString)
            except:
                output("""Sorry, I couldn't find that server.  Perhaps you
                misspelled it?  Also, be sure not to put the port in the
                server's name -- we'll ask you about that later.""")
        else:
            ip = 'no network available'

    output("""Found %s (%s).""" % (serverString, ip))

    # conf.supybot.networks.<network>.ssl
    output("""Most networks allow you to use a secure connection via SSL.
    If you are not sure whether this network supports SSL or not, check
    its website.""")
    use_ssl = yn('Do you want to use an SSL connection?', default=True)
    network.ssl.setValue(use_ssl)

    output("""IRC servers almost always accept connections on port
    6697 (or 6667 when not using SSL).  They can, however, accept connections
    anywhere their admins feel like they want to accept connections from.""")
    if yn('Does this server require connection on a non-standard port?',
          default=False):
        port = 0
        while not port:
            port = something('What port is that?')
            try:
                i = int(port)
                if not (0 < i < 65536):
                    raise ValueError()
            except ValueError:
                output("""That's not a valid port.""")
                port = 0
    else:
        if network.ssl.value:
            port = 6697
        else:
            port = 6667
    server = ':'.join([serverString, str(port)])
    network.servers.setValue([server])

    # conf.supybot.nick
    # Force the user into specifying a nick if it didn't have one already
    while True:
        nick = something('What nick would you like your bot to use?',
                         default=None)
        try:
            conf.supybot.nick.set(nick)
            break
        except registry.InvalidRegistryValue:
            output("""That's not a valid nick.  Go ahead and pick another.""")

    # conf.supybot.networks.<network>.sasl.{username,password}
    output("""Most networks allow you to create an account, using services
    typically known as NickServ. If you already have one, you can configure
    the bot to use that account by proving your username and password
    (also known as "SASL plain").
    It is usually not a requirement.""")
    if yn('Do you want your bot to use a network account?', default=False):
        while True:
            username = something('What account name should the bot use?')
            password = getpass('What is the password of that account?')
            if not username:
                if yn('Missing username. Do you want to try again?'):
                    continue
            if not password:
                if yn('Missing password. Do you want to try again?'):
                    continue
            network.sasl.username.set(username)
            network.sasl.password.set(password)
            break

    # conf.supybot.user
    if advanced:
        output("""If you've ever done a /whois on a person, you know that IRC
        provides a way for users to show the world their full name.  What would
        you like your bot's full name to be?  If you don't care, just press
        enter and it'll be the same as your bot's nick.""")
        user = ''
        user = something('What would you like your bot\'s full name to be?',
                         default=nick)
        conf.supybot.user.set(user)
    # conf.supybot.ident (if advanced)
    defaultIdent = 'limnoria'
    if advanced:
        output("""IRC servers also allow you to set your ident, which they
        might need if they can't find your identd server.  What would you like
        your ident to be?  If you don't care, press enter and we'll use
        'limnoria'.  In fact, we prefer that you do this, because it provides
        free advertising for Supybot when users /whois your bot.  But, of
        course, it's your call.""")
        while True:
            ident = something('What would you like your bot\'s ident to be?',
                              default=defaultIdent)
            try:
                conf.supybot.ident.set(ident)
                break
            except registry.InvalidRegistryValue:
                output("""That was not a valid ident.  Go ahead and pick
                another.""")
    else:
        conf.supybot.ident.set(defaultIdent)

    # conf.supybot.networks.<network>.password
    output("""Some servers require a password to connect to them.  Most
    public servers don't.  If you try to connect to a server and for some
    reason it just won't work, it might be that you need to set a
    password.""")
    if yn('Do you want to set such a password?', default=False):
        network.password.set(getpass())

    # conf.supybot.networks.<network>.channels
    output("""Of course, having an IRC bot isn't the most useful thing in the
    world unless you can make that bot join some channels.""")
    if yn('Do you want your bot to join some channels when it connects?',
          default=True):
        defaultChannels = ' '.join(network.channels())
        output("""Separate channels with spaces.  If the channel is locked
                  with a key, follow the channel name with the key separated
                  by a comma. For example:
                  #supybot-bots #mychannel,mykey #otherchannel""");
        while True:
            channels_input = something('What channels?', default=defaultChannels)
            channels = []
            error_ = False
            for channel in channels_input.split():
                L = channel.split(',')
                if len(L) == 0:
                    continue
                elif len(L) == 1:
                    (channel,) = L
                    channels.append(channel)
                elif len(L) == 2:
                    (channel, key) = L
                    channels.append(channel)
                    network.channels.key.get(channel).setValue(key)
                else:
                    output("""Too many commas in %s.""" % channel)
                    error_ = True
                    break
            if error_:
                break
            try:
                network.channels.set(' '.join(channels))
                break
            except registry.InvalidRegistryValue as e:
                output(""""%s" is an invalid IRC channel.  Be sure to prefix
                the channel with # (or +, or !, or &, but no one uses those
                channels, really).  Be sure the channel key (if you are
                supplying one) does not contain a comma.""" % e.channel)
    else:
        network.channels.setValue([])

    ###
    # Plugins
    ###
    def configurePlugin(module, advanced):
        if hasattr(module, 'configure'):
            output("""Beginning configuration for %s...""" %
                   module.Class.__name__)
            module.configure(advanced)
            print() # Blank line :)
            output("""Done!""")
        else:
            conf.registerPlugin(module.__name__, currentValue=True)

    plugins = getPlugins(pluginDirs + [plugin._pluginsDir])
    for s in ('Admin', 'User', 'Channel', 'Misc', 'Config', 'Utilities'):
        m = loadPlugin(s)
        if m is not None:
            configurePlugin(m, advanced)
        else:
            error('There was an error loading one of the core plugins that '
                  'under almost all circumstances are loaded.  Go ahead and '
                  'fix that error and run this script again.')
    clearLoadedPlugins(plugins, conf.supybot.plugins)

    output("""Now we're going to run you through plugin configuration. There's
           a variety of plugins in supybot by default, but you can create and
           add your own, of course. We'll allow you to take a look at the known
           plugins' descriptions and configure them
           if you like what you see.""")

    # bulk
    addedBulk = False
    if advanced and yn('Would you like to add plugins en masse first?'):
        addedBulk = True
        output(format("""The available plugins are: %L.""", plugins))
        output("""What plugins would you like to add?  If you've changed your
        mind and would rather not add plugins in bulk like this, just press
        enter and we'll move on to the individual plugin configuration.
        We suggest you to add Aka, Ctcp, Later, Network, Plugin, String,
        and Utilities""")
        massPlugins = anything('Separate plugin names by spaces or commas:')
        for name in re.split(r',?\s+', massPlugins):
            module = loadPlugin(name)
            if module is not None:
                configurePlugin(module, advanced)
                clearLoadedPlugins(plugins, conf.supybot.plugins)

    # individual
    if yn('Would you like to look at plugins individually?'):
        output("""Next comes your opportunity to learn more about the plugins
        that are available and select some (or all!) of them to run in your
        bot.  Before you have to make a decision, of course, you'll be able to
        see a short description of the plugin and, if you choose, an example
        session with the plugin.  Let's begin.""")
        # until we get example strings again, this will default to false
        #showUsage =yn('Would you like the option of seeing usage examples?')
        showUsage = False
        name = expect('What plugin would you like to look at?',
                      plugins, acceptEmpty=True)
        while name:
            module = loadPlugin(name)
            if module is not None:
                describePlugin(module, showUsage)
                if yn('Would you like to load this plugin?', default=True):
                    configurePlugin(module, advanced)
                    clearLoadedPlugins(plugins, conf.supybot.plugins)
            if not yn('Would you like add another plugin?'):
                break
            name = expect('What plugin would you like to look at?', plugins)

    ###
    # Sundry
    ###
    output("""Although supybot offers a supybot-adduser script, with which
    you can add users to your bot's user database, it's *very* important that
    you have an owner user for you bot.""")
    if yn('Would you like to add an owner user for your bot?', default=True):
        import supybot.ircdb as ircdb
        name = something('What should the owner\'s username be?')
        try:
            id = ircdb.users.getUserId(name)
            u = ircdb.users.getUser(id)
            if u._checkCapability('owner'):
                output("""That user already exists, and has owner capabilities
                already.  Perhaps you added it before? """)
                if yn('Do you want to remove its owner capability?',
                      default=False):
                    u.removeCapability('owner')
                    ircdb.users.setUser(u)
            else:
                output("""That user already exists, but doesn't have owner
                capabilities.""")
                if yn('Do you want to add to it owner capabilities?',
                      default=False):
                    u.addCapability('owner')
                    ircdb.users.setUser(u)
        except KeyError:
            password = getpass('What should the owner\'s password be?')
            u = ircdb.users.newUser()
            u.name = name
            u.setPassword(password)
            u.addCapability('owner')
            ircdb.users.setUser(u)

    output("""Of course, when you're in an IRC channel you can address the bot
    by its nick and it will respond, if you give it a valid command (it may or
    may not respond, depending on what your config variable replyWhenNotCommand
    is set to).  But your bot can also respond to a short "prefix character,"
    so instead of saying "bot: do this," you can say, "@do this" and achieve
    the same effect.  Of course, you don't *have* to have a prefix char, but
    if the bot ends up participating significantly in your channel, it'll ease
    things.""")
    if yn('Would you like to set the prefix char(s) for your bot?  ',
          default=True):
        output("""Enter any characters you want here, but be careful: they
        should be rare enough that people don't accidentally address the bot
        (simply because they'll probably be annoyed if they do address the bot
        on accident).  You can even have more than one.  I (jemfinch) am quite
        partial to @, but that's because I've been using it since my ocamlbot
        days.""")
        import supybot.callbacks as callbacks
        c = ''
        while not c:
            try:
                c = anything('What would you like your bot\'s prefix '
                             'character(s) to be?')
                conf.supybot.reply.whenAddressedBy.chars.set(c)
            except registry.InvalidRegistryValue as e:
                output(str(e))
                c = ''
    else:
        conf.supybot.reply.whenAddressedBy.chars.set('')

    ###
    # logging variables.
    ###

    if advanced:
        # conf.supybot.log.stdout
        output("""By default, your bot will log not only to files in the logs
        directory you gave it, but also to stdout.  We find this useful for
        debugging, and also just for the pretty output (it's colored!)""")
        stdout = not yn('Would you like to turn off this logging to stdout?',
                        default=False)
        conf.supybot.log.stdout.setValue(stdout)
        if conf.supybot.log.stdout():
            # conf.something
            output("""Some terminals may not be able to display the pretty
            colors logged to stderr.  By default, though, we turn the colors
            off for Windows machines and leave it on for *nix machines.""")
            if os.name != 'nt':
                conf.supybot.log.stdout.colorized.setValue(
                    not yn('Would you like to turn this colorization off?',
                    default=False))

        # conf.supybot.log.level
        output("""Your bot can handle debug messages at several priorities,
        CRITICAL, ERROR, WARNING, INFO, and DEBUG, in decreasing order of
        priority. By default, your bot will log all of these priorities except
        DEBUG.  You can, however, specify that it only log messages above a
        certain priority level.""")
        priority = str(conf.supybot.log.level)
        logLevel = something('What would you like the minimum priority to be?'
                             '  Just press enter to accept the default.',
                             default=priority).lower()
        while logLevel not in ['debug','info','warning','error','critical']:
            output("""That's not a valid priority.  Valid priorities include
            'DEBUG', 'INFO', 'WARNING', 'ERROR', and 'CRITICAL'""")
            logLevel = something('What would you like the minimum priority to '
                                 'be?  Just press enter to accept the default.',
                                 default=priority).lower()
        conf.supybot.log.level.set(logLevel)

        # conf.supybot.databases.plugins.channelSpecific

        output("""Many plugins in Supybot are channel-specific.  Their
        databases, likewise, are specific to each channel the bot is in.  Many
        people don't want this, so we have one central location in which to
        say that you would prefer all databases for all channels to be shared.
        This variable, supybot.databases.plugins.channelSpecific, is that
        place.""")

        conf.supybot.databases.plugins.channelSpecific.setValue(
            not yn('Would you like plugin databases to be shared by all '
                   'channels, rather than specific to each channel the '
                   'bot is in?'))

    output("""There are a lot of options we didn't ask you about simply
              because we'd rather you get up and running and have time
              left to play around with your bot.  But come back and see
              us!  When you've played around with your bot enough to
              know what you like, what you don't like, what you'd like
              to change, then take a look at your configuration file
              when your bot isn't running and read the comments,
              tweaking values to your heart's desire.""")

    # Let's make sure that src/ plugins are loaded.
    conf.registerPlugin('Admin', True)
    conf.registerPlugin('AutoMode', True)
    conf.registerPlugin('Channel', True)
    conf.registerPlugin('Config', True)
    conf.registerPlugin('Misc', True)
    conf.registerPlugin('Network', True)
    conf.registerPlugin('NickAuth', True)
    conf.registerPlugin('User', True)
    conf.registerPlugin('Utilities', True)

    ###
    # Write the registry
    ###

    # We're going to need to do a darcs predist thing here.
    #conf.supybot.debug.generated.setValue('...')

    if advanced:
        basedir = '.'
    filename = os.path.join(basedir, '%s.conf')

    filename = something("""In which file would you like to save
                         this config?""", default=filename % nick)
    if not filename.endswith('.conf'):
        filename += '.conf'
    registry.close(conf.supybot, os.path.expanduser(filename))

    # Done!
    output("""All done!  Your new bot configuration is %s.  If you're running
    a *nix based OS, you can probably start your bot with the command line
    "supybot %s".  If you're not running a *nix or similar machine, you'll
    just have to start it like you start all your other Python scripts.""" % \
                                                         (filename, filename))

def main():
    try:
        _main()
    except KeyboardInterrupt:
        # We may still be using bold text when exiting during a prompt
        if questions.useBold:
            import supybot.ansi as ansi
            print(ansi.RESET)
        print()
        print()
        output("""Well, it looks like you canceled out of the wizard before
        it was done.  Unfortunately, I didn't get to write anything to file.
        Please run the wizard again to completion.""")


if __name__ == '__main__':
    main()

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2015-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

###

import re
import sys
import time
import uuid
import functools

import supybot.gpg as gpg
import supybot.conf as conf
import supybot.utils as utils
import supybot.ircdb as ircdb
from supybot.commands import *
import supybot.plugins as plugins
import supybot.commands as commands
import supybot.ircutils as ircutils
import supybot.callbacks as callbacks
import http.client as http_client
try:
    from supybot.i18n import PluginInternationalization
    _ = PluginInternationalization('GPG')
except ImportError:
    # Placeholder that allows to run the plugin on a bot
    # without the i18n module
    _ = lambda x: x

def check_gpg_available(f):
    if gpg.available:
        return f
    else:
        if not gpg.found_gnupg_lib:
            def newf(self, irc, *args):
                irc.error(_('gnupg features are not available because '
                    'the python-gnupg library is not installed.'))
        elif not gpg.found_gnupg_bin:
            def newf(self, irc, *args):
                irc.error(_('gnupg features are not available because '
                    'the gnupg executable is not installed.'))
        else:
            # This case should never happen.
            def newf(self, irc, *args):
                irc.error(_('gnupg features are not available.'))
        newf.__doc__ = f.__doc__
        newf.__name__ = f.__name__
        return newf

if hasattr(http_client, '_MAXHEADERS'):
    safe_getUrl = utils.web.getUrl
else:
    def safe_getUrl(url):
        try:
            return commands.process(utils.web.getUrl, url,
                    timeout=10, heap_size=10*1024*1024,
                    pn='GPG')
        except (commands.ProcessTimeoutError, MemoryError):
            raise utils.web.Error(_('Page is too big or the server took '
                    'too much time to answer the request.'))

class GPG(callbacks.Plugin):
    """Provides authentication based on GPG keys.

    First you must associate your GPG key with your Limnoria account. The gpg
    add command takes two arguments, key id and key server.

    My key is 0x0C207F07B2F32B67 and it's on keyserver pool.sks-keyservers.net
    so and now I add it to my bot::

        <Mikaela> +gpg add 0x0C207F07B2F32B67 pool.sks-keyservers.net
        <Yvzabevn> 1 key imported, 0 unchanged, 0 not imported.

    Now I can get token to sign so I can identify::

        <Guest45020> +gpg gettoken
        <Yvzabevn> Your token is: {03640620-97ea-4fdf-b0c3-ce8fb62f2dc5}. Please sign it with your GPG key, paste it somewhere, and call the 'auth' command with the URL to the (raw) file containing the signature.

    Then I follow the instructions and sign my token in terminal::

        echo "{03640620-97ea-4fdf-b0c3-ce8fb62f2dc5}"|gpg --clearsign|curl -F 'sprunge=<-' http://sprunge.us

    Note that I sent the output to curl with flags to directly send the
    clearsigned content to sprunge.us pastebin. Curl should be installed on
    most of distributions and comes with msysgit. If you remove the curl part,
    you get the output to terminal and can pastebin it to any pastebin of
    your choice. Sprunge.us has only plain text and is easy so I used it in
    this example.

    And last I give the bot link to the plain text signature::

        <Guest45020> +gpg auth http://sprunge.us/DUdd
        <Yvzabevn> You are now authenticated as Mikaela.
    """
    class key(callbacks.Commands):
        @check_gpg_available
        def add(self, irc, msg, args, user, keyid, keyserver):
            """<key id> <key server>

            Add a GPG key to your account."""
            if keyid in user.gpgkeys:
                irc.error(_('This key is already associated with your '
                    'account.'))
                return
            result = gpg.keyring.recv_keys(keyserver, keyid)
            reply = format(_('%n imported, %i unchanged, %i not imported.'),
                    (result.imported, _('key')),
                    result.unchanged,
                    result.not_imported,
                    [x['fingerprint'] for x in result.results])
            if result.imported == 1:
                user.gpgkeys.append(keyid)
                irc.reply(reply)
            else:
                irc.error(reply)
        add = wrap(add, ['user',
                         ('somethingWithoutSpaces',
                             _('You must give a valid key id')),
                         ('somethingWithoutSpaces',
                             _('You must give a valid key server'))])

        @check_gpg_available
        def remove(self, irc, msg, args, user, fingerprint):
            """<fingerprint>

            Remove a GPG key from your account."""
            try:
                keyids = [x['keyid'] for x in gpg.keyring.list_keys()
                        if x['fingerprint'] == fingerprint]
                if len(keyids) == 0:
                    raise ValueError
                for keyid in keyids:
                    try:
                        user.gpgkeys.remove(keyid)
                    except ValueError:
                        user.gpgkeys.remove('0x' + keyid)
                gpg.keyring.delete_keys(fingerprint)
                irc.replySuccess()
            except ValueError:
                irc.error(_('GPG key not associated with your account.'))
        remove = wrap(remove, ['user', 'somethingWithoutSpaces'])

        @check_gpg_available
        def list(self, irc, msg, args, user):
            """takes no arguments

            List your GPG keys."""
            keyids = user.gpgkeys
            if len(keyids) == 0:
                irc.reply(_('No key is associated with your account.'))
            else:
                irc.reply(format('%L', keyids))
        list = wrap(list, ['user'])

    class signing(callbacks.Commands):
        def __init__(self, *args):
            super(GPG.signing, self).__init__(*args)
            self._tokens = {}

        def _expire_tokens(self):
            now = time.time()
            self._tokens = dict(filter(lambda x_y: x_y[1][1]>now,
                self._tokens.items()))

        @check_gpg_available
        def gettoken(self, irc, msg, args):
            """takes no arguments

            Send you a token that you'll have to sign with your key."""
            self._expire_tokens()
            token = '{%s}' % str(uuid.uuid4())
            lifetime = conf.supybot.plugins.GPG.auth.sign.TokenTimeout()
            self._tokens.update({token: (msg.prefix, time.time()+lifetime)})
            irc.reply(_('Your token is: %s. Please sign it with your '
                'GPG key, paste it somewhere, and call the \'auth\' '
                'command with the URL to the (raw) file containing the '
                'signature.') % token)
        gettoken = wrap(gettoken, [])

        _auth_re = re.compile(r'-----BEGIN PGP SIGNED MESSAGE-----\r?\n'
                r'Hash: .*\r?\n\r?\n'
                r'\s*({[0-9a-z-]+})\s*\r?\n'
                r'-----BEGIN PGP SIGNATURE-----\r?\n.*'
                r'\r?\n-----END PGP SIGNATURE-----',
                re.S)

        @check_gpg_available
        def auth(self, irc, msg, args, url):
            """<url>

            Check the GPG signature at the <url> and authenticates you if
            the key used is associated to a user."""
            self._expire_tokens()
            content = safe_getUrl(url)
            if isinstance(content, bytes):
                content = content.decode()
            match = self._auth_re.search(content)
            if not match:
                irc.error(_('Signature or token not found.'), Raise=True)
            data = match.group(0)
            token = match.group(1)
            if token not in self._tokens:
                irc.error(_('Unknown token. It may have expired before you '
                    'submit it.'), Raise=True)
            if self._tokens[token][0] != msg.prefix:
                irc.error(_('Your hostname/nick changed in the process. '
                    'Authentication aborted.'), Raise=True)
            verified = gpg.keyring.verify(data)
            if verified and verified.valid:
                keyid = verified.pubkey_fingerprint[-16:]
                prefix, expiry = self._tokens.pop(token)
                found = False
                for (id, user) in ircdb.users.items():
                    if keyid in [x[-len(keyid):] for x in user.gpgkeys]:
                        try:
                            user.addAuth(msg.prefix)
                        except ValueError:
                            irc.error(_('Your secure flag is true and your '
                                      'hostmask doesn\'t match any of your '
                                      'known hostmasks.'), Raise=True)
                        ircdb.users.setUser(user, flush=False)
                        irc.reply(_('You are now authenticated as %s.') %
                                user.name)
                        return
                irc.error(_('Unknown GPG key.'), Raise=True)
            else:
                irc.error(_('Signature could not be verified. Make sure '
                    'this is a valid GPG signature and the URL is valid.'))
        auth = wrap(auth, ['url'])


Class = GPG


# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2002-2004, Jeremiah Fincher
# Copyright (c) 2008-2010, James McCoy
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###
import json

import supybot.utils as utils
from supybot.commands import *
import supybot.callbacks as callbacks
from supybot.i18n import PluginInternationalization, internationalizeDocstring
from supybot.dynamicScope import dynamic
_ = PluginInternationalization('Google')

class Google(callbacks.Plugin):
    """
    This plugin provides access to Google services:

    1. translate

       Translates a string
       ``!translate en ar test``

    Check: `Supported language codes`_

    .. _Supported language codes: <https://cloud.google.com/translate/v2/using_rest#language-params>`
    """
    threaded = True

    def _translate(self, sourceLang, targetLang, text):
        headers = dict(utils.web.defaultHeaders)
        headers['User-agent'] = ('Mozilla/5.0 (X11; U; Linux i686) '
                                 'Gecko/20071127 Firefox/2.0.0.11')

        sourceLang = utils.web.urlquote(sourceLang)
        targetLang = utils.web.urlquote(targetLang)

        url = 'https://translate.googleapis.com/translate_a/single?' + \
            utils.web.urlencode({
                'client': 'gtx',
                'dt': 't',
                'sl': sourceLang,
                'tl': targetLang,
                'q': text})
        result = utils.web.getUrlFd(url, headers).read().decode('utf8')
        data = json.loads(result)

        try:
            language = data[2]
        except:
            language = 'unknown'

        if data[0]:
            return (''.join(x[0] for x in data[0]), language)
        else:
            return (_('No translations found.'), language)

    @internationalizeDocstring
    def translate(self, irc, msg, args, sourceLang, targetLang, text):
        """<source language> [to] <target language> <text>

        Returns <text> translated from <source language> into <target
        language>. <source language> and <target language> take language
        codes (not language names), which are listed here:
        https://cloud.google.com/translate/docs/languages
        """
        (text, language) = self._translate(sourceLang, targetLang, text)
        irc.reply(text, language)
    translate = wrap(translate, ['something', 'to', 'something', 'text'])

Class = Google


# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2005, Jeremiah Fincher
# Copyright (c) 2009, James McCoy
# Copyright (c) 2010-2024, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import re
import socket

import supybot.conf as conf
import supybot.utils as utils
from supybot.commands import *
import supybot.plugins as plugins
import supybot.commands as commands
import supybot.ircutils as ircutils
import supybot.callbacks as callbacks
from supybot.i18n import PluginInternationalization, internationalizeDocstring
_ = PluginInternationalization('Web')

from html.parser import HTMLParser
from html.entities import entitydefs
import http.client as http_client

class Title(utils.web.HtmlToText):
    entitydefs = entitydefs.copy()
    entitydefs['nbsp'] = ' '
    def __init__(self):
        self.inTitle = False
        self.inSvg = 0 # counter instead of boolean because svg can be nested
        utils.web.HtmlToText.__init__(self)

    @property
    def inHtmlTitle(self):
        return self.inTitle and self.inSvg == 0

    def handle_starttag(self, tag, attrs):
        if tag == 'title':
            self.inTitle = True
        elif tag == 'svg':
            self.inSvg += 1

    def handle_endtag(self, tag):
        if tag == 'title':
            self.inTitle = False
        elif tag == 'svg':
            self.inSvg = max(0, self.inSvg - 1)

    def append(self, data):
        if self.inHtmlTitle:
            super(Title, self).append(data)

class DelayedIrc:
    def __init__(self, irc):
        self._irc = irc
        self._replies = []
    def reply(self, *args, **kwargs):
        self._replies.append(('reply', args, kwargs))
    def error(self, *args, **kwargs):
        self._replies.append(('error', args, kwargs))
    def __getattr__(self, name):
        assert name not in ('reply', 'error', '_irc', '_msg', '_replies')
        return getattr(self._irc, name)

if hasattr(http_client, '_MAXHEADERS'):
    def fetch_sandbox(f):
        """Runs a command in a forked process with limited memory resources
        to prevent memory bomb caused by specially crafted http responses.

        On CPython versions with support for limiting the number of headers,
        this is the identity function"""
        return f
else:
    # For the following CPython versions (as well as the matching Pypy
    # versions):
    # * 2.6 before 2.6.9
    # * 2.7 before 2.7.9
    # * 3.2 before 3.2.6
    # * 3.3 before 3.3.3
    def fetch_sandbox(f):
        """Runs a command in a forked process with limited memory resources
        to prevent memory bomb caused by specially crafted http responses."""
        def process(self, irc, msg, *args, **kwargs):
            delayed_irc = DelayedIrc(irc)
            f(self, delayed_irc, msg, *args, **kwargs)
            return delayed_irc._replies
        def newf(self, irc, *args):
            try:
                replies = commands.process(process, self, irc, *args,
                        timeout=10, heap_size=10*1024*1024,
                        pn=self.name(), cn=f.__name__)
            except (commands.ProcessTimeoutError, MemoryError):
                raise utils.web.Error(_('Page is too big or the server took '
                        'too much time to answer the request.'))
            else:
                for (method, args, kwargs) in replies:
                    getattr(irc, method)(*args, **kwargs)
        newf.__doc__ = f.__doc__
        return newf

def catch_web_errors(f):
    """Display a nice error instead of "An error has occurred"."""
    def newf(self, irc, *args, **kwargs):
        try:
            f(self, irc, *args, **kwargs)
        except utils.web.Error as e:
            irc.reply(str(e))
    return utils.python.changeFunctionName(newf, f.__name__, f.__doc__)

class Web(callbacks.PluginRegexp):
    """Add the help for 'help Web' here."""
    regexps = ['titleSnarfer']
    threaded = True

    def noIgnore(self, irc, msg):
        return not self.registryValue('checkIgnored', msg.channel, irc.network)

    def getTitle(self, irc, url, raiseErrors, msg):
        size = conf.supybot.protocols.http.peekSize()

        def url_workaround(url):
            """Returns a new URL that should be the target of a new request,
            or None if the request is fine as it is.

            The returned URL may be the same as the parameter, in case
            something else was changed by this function through side-effects.
            """
            nonlocal size
            parsed_url = utils.web.urlparse(url)
            if parsed_url.netloc in ('youtube.com', 'youtu.be') \
                    or parsed_url.netloc.endswith(('.youtube.com')):
                # there is a lot of Javascript before the <title>
                if size < 819200:
                    size = max(819200, size)
                    return url
                else:
                    return None
            if parsed_url.netloc in ('reddit.com', 'www.reddit.com', 'new.reddit.com'):
                # Since 2022-03, New Reddit has 'Reddit - Dive into anything' as
                # <title> on every page.
                parsed_url = parsed_url._replace(netloc='old.reddit.com')
                url = utils.web.urlunparse(parsed_url)
                self.log.debug("Rewrite URL to %s", url)
                return url

            return None

        url = url_workaround(url) or url
        timeout = self.registryValue('timeout')
        headers = conf.defaultHttpHeaders(irc.network, msg.channel)
        try:
            fd = utils.web.getUrlFd(url, timeout=timeout, headers=headers)
            target = fd.geturl()
            fixed_target = url_workaround(target)
            if fixed_target is not None:
                # happens when using minification services linking to one of
                # the websites handled by url_workaround; eg. v.redd.it
                fd.close()
                fd = utils.web.getUrlFd(fixed_target, timeout=timeout, headers=headers)
                target = fd.geturl()
            text = fd.read(size)
            response_headers = fd.headers
            fd.close()
        except socket.timeout:
            if raiseErrors:
                irc.error(_('Connection to %s timed out') % url, Raise=True)
            else:
                self.log.info('Web plugins TitleSnarfer: URL <%s> timed out',
                              url)
                return
        except Exception as e:
            if raiseErrors:
                irc.error(_('That URL raised <' + str(e)) + '>',
                          Raise=True)
            else:
                self.log.info('Web plugin TitleSnarfer: URL <%s> raised <%s>',
                              url, str(e))
                return

        encoding = None
        if 'Content-Type' in fd.headers:
            # using p.partition('=') instead of 'p.split('=', 1)' because,
            # unlike RFC 7231, RFC 9110 allows an empty parameter list
            # after ';':
            # * https://www.rfc-editor.org/rfc/rfc9110.html#name-media-type
            # * https://www.rfc-editor.org/rfc/rfc9110.html#parameter
            mime_params = [p.partition('=')
                for p in fd.headers['Content-Type'].split(';')[1:]]
            mime_params = {k.strip(): v.strip() for (k, sep, v) in mime_params}
            if mime_params.get('charset'):
                encoding = mime_params['charset']

        encoding = encoding or utils.web.getEncoding(text) or 'utf8'

        try:
            text = text.decode(encoding, 'replace')
        except UnicodeDecodeError:
            if raiseErrors:
                irc.error(_('Could not guess the page\'s encoding. (Try '
                            'installing python-charade.)'), Raise=True)
            else:
                self.log.info('Web plugin TitleSnarfer: URL <%s> Could '
                              'not guess the page\'s encoding. (Try '
                              'installing python-charade.)', url)
                return
        try:
            parser = Title()
            parser.feed(text)
        except UnicodeDecodeError:
            # Workaround for Python 2
            # https://github.com/progval/Limnoria/issues/1359
            parser = Title()
            parser.feed(text.encode('utf8'))
        parser.close()
        title = utils.str.normalizeWhitespace(''.join(parser.data).strip())
        if title:
            return (target, title)
        elif raiseErrors:
            if len(text) < size:
                irc.error(_('That URL appears to have no HTML title.'),
                        Raise=True)
            else:
                irc.error(format(_('That URL appears to have no HTML title '
                                 'within the first %S.'), size), Raise=True)
        else:
            if len(text) < size:
                self.log.debug('Web plugin TitleSnarfer: URL <%s> appears '
                               'to have no HTML title. ', url)
            else:
                self.log.debug('Web plugin TitleSnarfer: URL <%s> appears '
                               'to have no HTML title within the first %S.',
                               url, size)

    @fetch_sandbox
    def titleSnarfer(self, irc, msg, match):
        channel = msg.channel
        network = irc.network
        if not channel:
            return
        if callbacks.addressed(irc, msg):
            return
        if self.registryValue('titleSnarfer', channel, network):
            url = match.group(0)
            if not self._checkURLWhitelist(irc, msg, url):
                return
            r = self.registryValue('nonSnarfingRegexp', channel, network)
            if r and r.search(url):
                self.log.debug('Not titleSnarfing %q.', url)
                return
            r = self.getTitle(irc, url, False, msg)
            if not r:
                return
            (target, title) = r
            if title:
                domain = utils.web.getDomain(target
                        if self.registryValue('snarferShowTargetDomain',
                                              channel, network)
                        else url)
                prefix = self.registryValue('snarferPrefix', channel, network)
                if prefix:
                    s = "%s %s" % (prefix, title)
                else:
                    s = title
                if self.registryValue('snarferShowDomain', channel, network):
                    s += format(_(' (at %s)'), domain)
                irc.reply(s, prefixNick=False)
        if self.registryValue('snarfMultipleUrls', channel, network):
            # FIXME: hack
            msg.tag('repliedTo', False)
    titleSnarfer = urlSnarfer(titleSnarfer)
    titleSnarfer.__doc__ = utils.web._httpUrlRe

    def _checkURLWhitelist(self, irc, msg, url):
        if not self.registryValue('urlWhitelist',
                                  channel=msg.channel, network=irc.network):
            return True
        passed = False
        for wu in self.registryValue('urlWhitelist',
                                     channel=msg.channel, network=irc.network):
            if wu.endswith('/') and url.find(wu) == 0:
                passed = True
                break
            if (not wu.endswith('/')) and (url.find(wu + '/') == 0 or url == wu):
                passed = True
                break
        return passed

    @wrap(['httpUrl'])
    @catch_web_errors
    @fetch_sandbox
    def headers(self, irc, msg, args, url):
        """<url>

        Returns the HTTP headers of <url>.  Only HTTP urls are valid, of
        course.
        """
        if not self._checkURLWhitelist(irc, msg, url):
            irc.error("This url is not on the whitelist.")
            return
        timeout = self.registryValue('timeout')
        fd = utils.web.getUrlFd(url, timeout=timeout)
        try:
            s = ', '.join([format(_('%s: %s'), k, v)
                           for (k, v) in fd.headers.items()])
            irc.reply(s)
        finally:
            fd.close()

    @wrap(['httpUrl'])
    @catch_web_errors
    @fetch_sandbox
    def location(self, irc, msg, args, url):
        """<url>

        If the <url> is redirected to another page, returns the URL of that
        page. This works even if there are multiple redirects.
        Only HTTP urls are valid.
        Useful to "un-tinify" URLs."""
        timeout = self.registryValue('timeout')
        (target, text) = utils.web.getUrlTargetAndContent(url, size=60,
            timeout=timeout)
        irc.reply(target)

    _doctypeRe = re.compile(r'(<!DOCTYPE[^>]+>)', re.M)
    @wrap(['httpUrl'])
    @catch_web_errors
    @fetch_sandbox
    def doctype(self, irc, msg, args, url):
        """<url>

        Returns the DOCTYPE string of <url>.  Only HTTP urls are valid, of
        course.
        """
        if not self._checkURLWhitelist(irc, msg, url):
            irc.error("This url is not on the whitelist.")
            return
        size = conf.supybot.protocols.http.peekSize()
        timeout = self.registryValue('timeout')
        s = utils.web.getUrl(url, size=size, timeout=timeout).decode('utf8')
        m = self._doctypeRe.search(s)
        if m:
            s = utils.str.normalizeWhitespace(m.group(0))
            irc.reply(s)
        else:
            irc.reply(_('That URL has no specified doctype.'))

    @wrap(['httpUrl'])
    @catch_web_errors
    @fetch_sandbox
    def size(self, irc, msg, args, url):
        """<url>

        Returns the Content-Length header of <url>.  Only HTTP urls are valid,
        of course.
        """
        if not self._checkURLWhitelist(irc, msg, url):
            irc.error("This url is not on the whitelist.")
            return
        timeout = self.registryValue('timeout')
        fd = utils.web.getUrlFd(url, timeout=timeout)
        try:
            try:
                size = fd.headers['Content-Length']
                if size is None:
                    raise KeyError('content-length')
                irc.reply(format(_('%u is %S long.'), url, int(size)))
            except KeyError:
                size = conf.supybot.protocols.http.peekSize()
                s = fd.read(size)
                if len(s) != size:
                    irc.reply(format(_('%u is %S long.'), url, len(s)))
                else:
                    irc.reply(format(_('The server didn\'t tell me how long %u '
                                     'is but it\'s longer than %S.'),
                                     url, size))
        finally:
            fd.close()

    @wrap([getopts({'no-filter': ''}), 'httpUrl'])
    @catch_web_errors
    @fetch_sandbox
    def title(self, irc, msg, args, optlist, url):
        """[--no-filter] <url>

        Returns the HTML <title>...</title> of a URL.
        If --no-filter is given, the bot won't strip special chars (action,
        DCC, ...).
        """
        if not self._checkURLWhitelist(irc, msg, url):
            irc.error("This url is not on the whitelist.")
            return
        r = self.getTitle(irc, url, True, msg)
        if not r:
            return
        (target, title) = r
        if title:
            if not [y for x,y in optlist if x == 'no-filter']:
                for i in range(1, 4):
                    title = title.replace(chr(i), '')
            irc.reply(title)

    @wrap(['text'])
    def urlquote(self, irc, msg, args, text):
        """<text>

        Returns the URL quoted form of the text.
        """
        irc.reply(utils.web.urlquote(text))

    @wrap(['text'])
    def urlunquote(self, irc, msg, args, text):
        """<text>

        Returns the text un-URL quoted.
        """
        s = utils.web.urlunquote(text)
        irc.reply(s)

    @wrap(['url'])
    @catch_web_errors
    @fetch_sandbox
    def fetch(self, irc, msg, args, url):
        """<url>

        Returns the contents of <url>, or as much as is configured in
        supybot.plugins.Web.fetch.maximum.  If that configuration variable is
        set to 0, this command will be effectively disabled.
        """
        if not self._checkURLWhitelist(irc, msg, url):
            irc.error("This url is not on the whitelist.")
            return
        max = self.registryValue('fetch.maximum')
        timeout = self.registryValue('fetch.timeout')
        if not max:
            irc.error(_('This command is disabled '
                      '(supybot.plugins.Web.fetch.maximum is set to 0).'),
                      Raise=True)
        fd = utils.web.getUrl(url, size=max, timeout=timeout).decode('utf8')
        irc.reply(fd)

Class = Web

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:
